/*
 CognitiveEnv.java
 Clement Jonquet - PhD student in computer science - LIRMM - July 2005
*/

package madkit.strobekit;

import java.util.Vector;

import gnu.mapping.Environment;
import gnu.mapping.Location;
import gnu.mapping.LocationEnumeration;
import gnu.mapping.SimpleEnvironment;
import kawa.standard.Scheme;

import madkit.kernel.AgentAddress;

/**
 * This class define the cognitive environment structure of the STROBEKIT API. A <code>CognitiveEnv</code> belongs to a Strobe agent and is dedicated to another one. It has two input and output stream of messages.
 * <p> A <code>CognitiveEnv</code> is associated to a Scheme environment (provided by Kawa). This Scheme environment contains a Scheme interpreter used by the CognitiveEnv to evaluate messages. 
 * <p>
 * This class contains methods for: 
 * <ul>
 * 	<li> input and output streams management,
 * 	<li> Scheme environment management,
 * 	<li> evaluation of messages,
 *  <li> constructing Scheme environment by copy.
 * </ul>
 *
 * @author Clement Jonquet - Mail to <a href="mailto:jonquet@lirmm.fr">jonquet@lirmm.fr</a>
 * @version 0.1	01/30/2006
 */

public class CognitiveEnv{

//	-----------	
//	Attributes
//	-----------
	
	// OWNERAGENT, DEDICATEDAGENT - The agents owner and whose the CoEnv is dedicated to 
    protected StrobeAbstractAgent ownerAgent;
	protected AgentAddress dedicatedAgent;
	
    // INPUTSTREAM, OUTPUTSTREAM - Messages streams
	protected Vector inputStream = new Vector();
	protected Vector outputStream = new Vector();

	// BIND - The Scheme environment structure
	SimpleEnvironment bind;
		
	// INIT-EVAL
	private static String STROBEMSGFILE="C:/Documents and Settings/Clement-Travail/Mes documents/jonquet/MadKit/MadKit4.0.9/plugins/strobe/strobemsg.scm";
	
//	--------
//  Methods
//	--------	
	
    //	-------------	
    //  Constructors
    //	-------------	
 	
	/**  Create a new cognitive environment dedicated to <code>dedicated</code> which belongs to <code>owner</code>.*/
    public CognitiveEnv(StrobeAbstractAgent owner, AgentAddress dedicated){
		ownerAgent = owner;
		dedicatedAgent = dedicated;
		bind = copyBind(ownerAgent.initBind);
		initializeWithSchemeFile(STROBEMSGFILE);
	}
	
    /**  Create a new cognitive environment (by coping the Scheme environment <code>scmenv</code>) dedicated to <code>dedicated</code> which belongs to <code>owner</code>.*/
    CognitiveEnv(StrobeAbstractAgent owner, AgentAddress dedicated, SimpleEnvironment scmenv){
		ownerAgent = owner;
		dedicatedAgent = dedicated;
		bind = copyBind(scmenv);
	}
    
    /** Allow to load a Scheme file in the cognitive environment.*/
    public void initializeWithSchemeFile(String scmfilename){
		String exp = "(load \""+scmfilename+"\")";
		evaluate(exp);
	}

	//	------------------	
	//  Accessors
	//	------------------	
    
    /** Return the cognitive environment Scheme environment.*/
    public SimpleEnvironment getBind() {
		return bind;
	}
	    
	//	-----------------------------	
    //  Input/Ouput stream processing
    //	-----------------------------
	
    /** Return the cognitive environment input stream of messages.*/
    public Vector getInputStream() {return inputStream;}

	/** Return the cognitive environment output stream of messages.*/
    public Vector getOutputStream() {return outputStream;}

	/** Add a message to a cognitive environment stream (input or output).*/
    public void pushStream(Vector msgStream, StrobeMsg msg){
		msgStream.add(msg);
	}
	
    /** Remove a message to a cognitive environment stream (input or output).*/
    public StrobeMsg popStream(Vector msgStream){
		return (StrobeMsg)msgStream.remove(msgStream.size()-1);
	}
	
    /** Return <code>true</code> if <code>msgStream</code> is empty*/ 
    public boolean isStreamEmpty(Vector msgStream){
		return msgStream.isEmpty();
	}
		
	//	---------------------	
    //  idialogue processing
    //	---------------------

    /*
	public boolean isIdialogueDone(){
		String result = evaluate("idialogueDone");
		if(result.equals("#f")){
			return false;
		}
		else return true;
	}

	public void setIdialogueDone(boolean iD) {
		if(iD){
			evaluate("(set! idialogueDone #t)");
		}
		else evaluate("set! idialogueDone #f)");
	}
	
	public void stepFcn(){
		
	}
	*/
    
	//	-----------	
    //  Evaluation
    //	-----------	

	private String evaluate(String exp){
		boolean error = false;
		Object result = "error";
		
		try{
			//ownerAgent.printlnOut("evaluate: "+exp);
			result = Scheme.eval(exp, bind);	
		}
		catch(java.lang.Throwable e){
			error = true;
		}
		if(error){
			ownerAgent.println("Error during evaluation");
		}
		
		return result.toString();
	}
	
	/** Evaluate a Strobe message in the cognitive environment (with the local interpreter); if the result is a message then it is push in the cognitive environment output stream else, the method <code>analyzeAnswer</code> of the owner Strobe agent is called.*/ 
    public void evaluateStrobeMsg(StrobeMsg msg){
		//String content = msg.getContent();
		String result;
		String exp;
		StrobeMsg answerMsg;
		
		ownerAgent.println("Message's evaluation...");
		exp = "(eval '"+msg.toSchemeExp()+")";
		//exp = "(eval '"+content+")";
		result = evaluate(exp);	
		ownerAgent.println("Result: "+result);
		if(isResultMsg(result)){
			answerMsg = stringToStrobeMsg(result);
			pushStream(outputStream, answerMsg);
		}
		else{
			ownerAgent.analyzeAnswer(result);
		}	
	}
	
	private boolean isResultMsg(String res){
		int i = res.indexOf(" ");
		if(i>0) {
			res = res.substring(0, i);
		}
		//ownerAgent.printlnOut("res: "+res);
		return res.equals("(strobemsg");
	}
	
	private StrobeMsg stringToStrobeMsg(String stringMsg){
		// StrobeMsg Scheme format (strobemsg sender receiver performative content)
		StrobeMsg msg = null;
		boolean error = false;
		String perfo, content;
		int i;
		try{
			i = stringMsg.indexOf(" ");
			stringMsg = stringMsg.substring(i+1, stringMsg.length());
			
			i = stringMsg.indexOf(" ");
			stringMsg = stringMsg.substring(i+1, stringMsg.length());
			
			i = stringMsg.indexOf(" ");
			stringMsg = stringMsg.substring(i+1, stringMsg.length());
			
			i = stringMsg.indexOf(" ");
			perfo =  stringMsg.substring(0, i);
			content =  stringMsg.substring(i+1, stringMsg.length()-1);
			//ownerAgent.printlnOut("perfo: "+perfo+ ", content: "+ content);
			msg = new StrobeMsg(perfo, content);
		}
		catch(java.lang.Throwable e){
			error = true;
		}
		if(error){
			ownerAgent.println("Error while processing string in stringToStrobeMsg");
			msg = new StrobeMsg("error", "error");		 
		}
		return msg;
	}
	
	//	------------------	
    //  Copy construtor
    //	------------------	

	/** Return a Scheme environment (class <code>SimpleEnvironment</code>) constructed by copy of <code>senv</code>*/
	static SimpleEnvironment copyBind(SimpleEnvironment senv){
		SimpleEnvironment copiedBind = Environment.make("zeo", senv);
		copiedBind.makeShared();
		for (LocationEnumeration b = senv.enumerateLocations(); b.hasMoreElements();){
			Location l = b.nextLocation();
			copiedBind.define(l.getKeySymbol(),l.getKeyProperty(),l.getValue());
		}
		copiedBind.makeShared();
		return copiedBind;
	}
	
	

		

	
	
}	