package madkit.strobekit;

import gnu.mapping.CharArrayOutPort;
import gnu.mapping.Environment;
import gnu.mapping.OutPort;
import gnu.mapping.SimpleEnvironment;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import kawa.standard.Scheme;
import madkit.kernel.Agent;
import madkit.kernel.AgentAddress;

/**
 * The main class of the STROBEKIT agent API.
 * Strobe agents are considered to inherit from this abstract class.
 * <p>
 * This class contains methods for: 
 * <ul>
 * 	<li> Cognitive Environment (class <code>CogintiveEnv</code>) management
 * 	<li> Strobe Message (class <code>StrobeMsg</code>) management
 * 	<li> defining the agent REPL (Read - Eval - Print - Listen) loop
 * 	<li> 
 * </ul>
 *
 *<p>
 * Remarks concerning Strobe agents:
 * <ul>
 * 	<li> They are by default member of the community "strobe".
 * 	<li> They have at least one by default group and role.
 *  <li> They have a set of <code>CognitiveEnv</code>: One global and private called <code>globalEnv</code>, and a set of local ones (dedicated to each acquaintance) represented by a <code>hashmap</code> of pair (<code>AgentAddress</code>, <code>CognitiveEnv</code>).
 *  <li> "Viewed" agents are all the agents a Strobe agent see and can communicate with. 
 *  <li> An "Acquaintance" is a an agent for which a Strobe agent has a dedicated local cognitive environment; As soon as at least a message was excahnged between a Strobe agent and another one, it becomes an acquaintance.  
 *  <li>
 *  <li>
 *  <li> 
 * </ul>
 * 
 * 
 * @author Clement Jonquet - Mail to <a href="mailto:jonquet@lirmm.fr">jonquet@lirmm.fr</a>
 * @version 0.1	01/30/2006
 */


public abstract class StrobeAbstractAgent extends Agent {

//	-----------	
//	Attributes
//	-----------
	
	// MEN - Mental states instance
    //MentalStates mental = new MentalStates(this);
	
	// GROUPE, ROLE - A stroba has (for the moment) only ONE role and ONE group
    protected String group,role;
    
	// COENVS - Set of pair (agentAdress - CognitiveEnv) representing the set of local cognitive environments
    protected HashMap localEnvs;
    // voir meme transformer cette hashmap en un autre objet que je m'embette une fois afaire l'API et apres j'utilise que l'API
    
    // GLOBALENV - The agent global cognitive environment
    protected CognitiveEnv globalEnv; 
    
    // SCM - A Scheme object necessary in order to create Environment
    protected Scheme scm;
    
    // OPORT - A Scheme OutPort uses to recover Scheme evaluation outputs
    //public OutPort oPort; 
    
    // INITBIND - A SimpleEnvironment that will be used for CognitiveEnv's bind
    SimpleEnvironment initBind;
    private static String INITEVALFILE = "C:/Documents and Settings/Clement-Travail/Mes documents/jonquet/MadKit/MadKit4.0.9/plugins/strobekit/initeval.scm";
    
    // ERRORFILE - A file which is used to print the Scheme display messages
    String errorFile;
    private static String ERRORFILEPATH = "C:/Documents and Settings/Clement-Travail/Mes documents/jonquet/MadKit/MadKit4.0.9/plugins/strobekit/error-files/";
    	
//	--------
//  Methods
//	--------	
	
    //	-------------	
    //  Constructors
    //	-------------	
 
    private void genericSetUp(){
    	
    	// localEnvs creation
    	localEnvs = new HashMap();
    	
    	// Scheme object initialization
    	scm = new Scheme();
    	//OutPort.setOutDefault((OutPort)new CharArrayOutPort());
    	Scheme.registerEnvironment();
    	scm = Scheme.getInstance();
    	
    	// initBind initialization
    	initBind = CognitiveEnv.copyBind((SimpleEnvironment)Environment.getCurrent());
    }
    
    /**  Default no-args constructor, which just set-up internal structures.*/
    public StrobeAbstractAgent(){		
    	genericSetUp();
    }
	
    /**  Do the same that StrobeAbstractAgent() but set-up also agent name, group, and role.*/
    public StrobeAbstractAgent(String n, String g, String r){
       	genericSetUp();
       	
    	// Name, group and role setting
    	setName(n);
    	group = g;
    	role = r;
    }
    
	//	----------	
	//  Accessors
	//	----------	

    /**  Return the by default agent group.*/
	public String getStrobeAgentGroup(){return group;}
	public void setStrobeAgentGroup(String g){this.group = g;}
    
	/** Return the by default agent role.*/
	public String getStrobeAgentRole(){return role;}
	public void setStrobeAgentRole(String r){this.role = r;}
    
	/** Return the agent global cognitive environment.*/
	public CognitiveEnv getGlobalEnv(){return globalEnv;}
	public void setGlobalEnv(CognitiveEnv gE) {this.globalEnv = gE;}

	/** Return the agent local environments set; an <code>hashmap</code> of pair (<code>AgentAddress</code>, <code>CognitiveEnv</code>).*/
	public HashMap getLocalEnvs(){return localEnvs;}
	public void setLocalEnvs(HashMap lEs) {this.localEnvs = lEs;}

	//	---------------	
	//	MadKit methods
	//	---------------	

	/** Method initially called when the micro-kernel registers the agent; Strobe agents have to use the method <code>StrobeActivate</code>. */
	public final void activate(){
		
		// errorFile initialisation
    	errorFile = ERRORFILEPATH+getName()+".txt";
    	Scheme.eval("(define *error-file* \""+errorFile+"\")", initBind);
    	
    	// initBind initialisation (loading of the INITEVALFILE file)
    	initializeCognitiveEnvWith(INITEVALFILE);
    	
    	// globalEnv creation
    	globalEnv = new CognitiveEnv(this, getAddress());
    
    	strobeActivate();
	}
	
	/** This method defines the main behavior for threaded agents; Strobe agents have to use the method <code>StrobeLive</code>.*/
	public final void live(){
		strobeLive();
	}
	
	/** Method called by the agent micro-kernel at the end of the agent lifecycle; Strobe agents have to use the method <code>StrobeEnd</code>.*/
	public final void end(){
		strobeEnd();
		Scheme.eval("(close-output-port *error-port*)", initBind);
	}
	
	/** Replace the method <code>activate</code> for Strobe agents .*/
	public void strobeActivate(){}
	/** Replace the method <code>live</code> for Strobe agents .*/
	public void strobeLive(){}
	/** Replace the method <code>end</code> for Strobe agents .*/
	public void strobeEnd(){}
	
	//	-------------------------------------------	
	//	Viewed agents and acquaintances processing
	//	-------------------------------------------
    
	/** Return an array of agent addresses corresponding to all the agents (including itself) a Strobe agent can see i.e., having the same <code>groupe</code> and <code>role</code> in the "strobe" community.*/
	public AgentAddress[] viewedAgentAddress(){
		return getAgentsWithRole("strobe", group, role);
	}
	
	/** Return an array of agent names corresponding to all the agents (including itself) a Strobe agent can see i.e., having the same <code>groupe</code> and <code>role</code> in the "strobe" community.*/
	public String[] viewedAgentName(){
		AgentAddress[] myViewedAddress = viewedAgentAddress();
		String[] myViewName = new String[myViewedAddress.length];
		for(int i = 0; i < myViewedAddress.length; i++){
			if(myViewedAddress[i] != getAddress()) myViewName[i] = myViewedAddress[i].getName();
			else myViewName[i]="Myself";
	    }
		return myViewName;
	}
	
	/** Return an array of agent addresses corresponding to all the agents a Strobe agent has a cognitive environment dedicated to.*/
	public AgentAddress[] acquaintanceAddresses(){
		AgentAddress[] acqAddresses = new AgentAddress[localEnvs.size()];
		Object[] objectArray = localEnvs.keySet().toArray();
		for(int i = 0; i < localEnvs.size(); i++){
			acqAddresses[i] = (AgentAddress)objectArray[i];
			}
		return acqAddresses;
	}
		
	/** Return an array of agent names corresponding to all the agents a Strobe agent has a cognitive environment dedicated to.*/
	public String[] acquaintanceNames(){
		AgentAddress[] acqAddresses = acquaintanceAddresses();
		String[] acqNames = new String[acqAddresses.length];
		for(int i = 0; i < acqAddresses.length; i++){
			if(acqAddresses[i] != getAddress()) acqNames[i] = acqAddresses[i].getName();
			else acqNames[i]="Myself";
	    }
		return acqNames;
	}
	
	/** Return an array of agent addresses corresponding to itself (first position) and all the agents it has a cognitive environment dedicated to.*/
	public AgentAddress[] acquaintanceAndIAddresses(){
		AgentAddress[] acqAndIAddresses = new AgentAddress[localEnvs.size()+1];
		acqAndIAddresses[0] = getAddress();
		Object[] objectArray = localEnvs.keySet().toArray();
		for(int i = 0; i < localEnvs.size(); i++){
			acqAndIAddresses[i+1] = (AgentAddress)objectArray[i];
		}
		return acqAndIAddresses;
	}
	
	/** Return an array of agent namess corresponding to itself (first position) and all the agents it has a cognitive environment dedicated to.*/
	public String[] acquaintanceAndINames(){
		AgentAddress[] acqAddresses = acquaintanceAddresses();
		String[] acqAndINames = new String[acqAddresses.length+1];
		acqAndINames[0] = "Myself";
		for(int i = 0; i < acqAddresses.length; i++){
			acqAndINames[i+1] = acqAddresses[i].getName();
		}
		return acqAndINames;
	}

	/** Return <code>true</code> if <code>a</code> is already an acquaintance.*/
	public boolean isAlreadyAcquaintance(AgentAddress a){
		return localEnvs.containsKey(a);
	}

    //	----------------------	
	//	Adding acquaintances
	//	----------------------
   
	/** This method has to be redefined by Strobe agents; it corresponds to the adding acquaintance policy i.e., how a Strobe agents create a new <code>CognitiveEnv</code> when they meet a new agent for the very first time.*/
	public abstract void addAcquaintance(AgentAddress a);
    
	/** Create a new local cognitive environment for <code>a</code> by copying the Strobe agent global cognitive environment.*/
	public void addAcquaintanceByCopyingGlobalEnv(AgentAddress a){
    	addAcquaintanceByCopyingEnv(a, getAddress());
    }
    
	/** Create a new local cognitive environment for <code>newa</code> by copying the Strobe agent local cognitive environment dedicated to <code>olda</code>.*/
	public void addAcquaintanceByCopyingLocalEnv(AgentAddress newa, AgentAddress olda){
    	addAcquaintanceByCopyingEnv(newa, olda);
    }
 
	private void addAcquaintanceByCopyingEnv(AgentAddress newa, AgentAddress olda){
    	CognitiveEnv localEnv;
    	if (olda==getAddress()){
    		localEnv = globalEnv;
    	}
    	else {localEnv = (CognitiveEnv)localEnvs.get(olda);}
    	SimpleEnvironment env = localEnv.getBind();
    	CognitiveEnv newLocalEnv = new CognitiveEnv(this, newa, env);
    	localEnvs.put(newa, newLocalEnv);
    }
    
	/** Use the same local cognitive environment for <code>newa</code> by sharing the Strobe agent local cognitive environment dedicated to <code>olda</code>.*/
	public void addAcquaintanceBySharingLocalEnv(AgentAddress newa, AgentAddress olda){
    	CognitiveEnv newLocalEnv = (CognitiveEnv)localEnvs.get(olda);
		localEnvs.put(newa, newLocalEnv);
    }
	
	//	----------------------------------	
	//	Cognitive environments processing
	//	----------------------------------

	/** Load the file <code>scmfilename</code> in the by default Scheme environment used for constructing <code>CognitiveEnv</code> of this Strobe agent.*/
	public void initializeCognitiveEnvWith(String scmfilename){
		Scheme.eval("(load \""+scmfilename+"\")", initBind);
	}

	//	-----------------------	
	//	Organization processing
	//	-----------------------
    
	/** This method creates and/or joins a group corresponding to the bydefault Strobe agent group.*/
	public void createOrJoinGroup(String g){
    	print("My group is: "+g+".\n");
    	if (isGroup ("strobe",g)){
    		println("It is already existing... I join...");
	    }
    	else{
    		println("It doesn't exist. I create it.");
    		createGroup(true,"strobe",g, null, null);
	    }
	}
	
	//	----------------------	
	//	Message processing
	//	----------------------

	/** Print the <code>StrobeMsg</code> in the Strobe agent output writer just after sending it (using <code>sendMessage</code>).*/
	public void sendMessageAndPrint(AgentAddress a, StrobeMsg m){
		sendMessage(a, m);
		println("\n Msg sent: "+m.toString());
	}
	
	/** Choose (with Dynamic Scheduler) next message to interpret and return it; If the message sender is not already an acquaintance then <code>addAcquaintance</code> is called.*/
	public StrobeMsg read(){
    	StrobeMsg msg = dynamicScheduler();
	    println("\n Msg received: "+msg.toString());
	    //	If the sender is not already an acquaintance
    	if( msg.getSender()!=getAddress() && !isAlreadyAcquaintance(msg.getSender())){
    		addAcquaintance(msg.getSender());
		}
	    return msg;
    }
    
	/** Wait for and return the next message to interpret according to the Strobe agent policy (by default, called <code>waitNextMessage</code>).*/
	public StrobeMsg dynamicScheduler(){
    	//return (StrobeMsg)nextMessage();
    	return (StrobeMsg)waitNextMessage();
    }
           
    //non utilise pour le moment 
	/** Dispatch the Strobe message in the given local cognitive environment inputs stream.*/
	private void dispatchMessage(StrobeMsg msg){
		// If the sender is already an acquaintance
    	if(isAlreadyAcquaintance(msg.getSender())){
			CognitiveEnv localEnv = (CognitiveEnv)localEnvs.get(msg.getSender());
			localEnv.pushStream(localEnv.inputStream, msg);
		}
    	// If the sender is not an acquaintanced yet
		//...
    }
    
	/** Evaluate the Strobe message in the good cognitive environment.*/
	public void eval(StrobeMsg msg){
		CognitiveEnv localEnv;
		if(msg.getSender()!=getAddress()){
    		localEnv = (CognitiveEnv)localEnvs.get(msg.getSender());
		}
		else {localEnv = globalEnv;}
    	localEnv.evaluateStrobeMsg(msg);
    }
    
	/** Browse all the cognitive environments (locals and global) and send messages present in output streams.*/
	public void print(){
    	Set entries = localEnvs.entrySet();
    	Iterator it = entries.iterator();
    	while (it.hasNext()){
    		Map.Entry e = (Map.Entry)it.next();
    		CognitiveEnv localEnv=(CognitiveEnv)e.getValue();
    		if (!localEnv.isStreamEmpty(localEnv.outputStream)){
    			StrobeMsg msg = localEnv.popStream(localEnv.outputStream);
    			sendMessageAndPrint(localEnv.dedicatedAgent, msg);
    		}
    	}
    	if (!globalEnv.isStreamEmpty(globalEnv.outputStream)){
			StrobeMsg msg = globalEnv.popStream(globalEnv.outputStream);
			sendMessageAndPrint(getAddress(), msg);
		}
    }
	
	/** This method has to be redefined by Strobe agents; it corresponds to the way answer returned by Strobe message evaluation are analyzed. */
	public abstract void analyzeAnswer(String res);	
	
			
}
